//	Player.cpp - Class representing one of the players.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "drawing/Drawer.h"
#include "MiscHelper.h"
#include "Player.h"

//------------------------------------------------------------------------------
Player::Player(const TwoFloats& pos, float ang, const ThreeFloats& col):
Collidable(L"Player"),
position(pos),
colour(col),
angle(ang),
speed(4.0f),
powerLevel(0.0f),
armPos(0.0f),
collidePosition(position),
tailFlash(false),
tailFlashCount(1),
flameIndex(0),
flameCount(7),
newCol(1.0f, 0.75f, 0.45f),
flameInterp(0.0f),
upIsDown(false),
downIsDown(false),
leftIsDown(false),
rightIsDown(false)
{
	int i;

	if(colour == ThreeFloats(1.0f, 0.45f, 0.45f))
		oldCol = ThreeFloats(0.53f, 0.45f, 1.0f);
	else
		oldCol = ThreeFloats(1.0f, 0.45f, 0.45f);

	flameImages[0] = L"PlayerFlame";
	flameImages[1] = L"PlayerFlame1";
	flameImages[2] = L"PlayerFlame2";
	flameImages[3] = L"PlayerFlame3";
	flameImages[4] = L"PlayerFlame2";
	flameImages[5] = L"PlayerFlame1";

	flameOffsets[0] = 0.0f;
	flameOffsets[1] = 4.0f;
	flameOffsets[2] = 9.0f;
	flameOffsets[3] = 15.0f;
	flameOffsets[4] = 9.0f;
	flameOffsets[5] = 4.0f;

	for(i=0;i<StartingTailSize;++i)
	{
		tail.push_back(position);
		tempTail.push_front(position);
		tailSize.push_back(32.0f);
		tailCol.push_back(colour);
		tailAlpha.push_back(1.0f);
	}

	for(i=0;i<4;++i)
	{
		tailSize[i] = 0.0f;
		tailAlpha[i] = 0.0f;
	}
	for(i=4;i<12;++i)
	{
		tailSize[i] = fakeLog((float)(i-4)*0.125f)*32.0f;
		//tailAlpha[i] = (float)(i-4)*0.125f;
	}
}

//------------------------------------------------------------------------------
Player::~Player()
{

}

//------------------------------------------------------------------------------
void Player::update()
{
	unsigned int i;
	float lastArm;

	if(collidePosition != position)
		position = collidePosition;

	if(direction != desiredDirection)
	{
		const TwoFloats origin;

		if(desiredDirection != 0.0f)
			direction += (desiredDirection-direction) * 0.2f;
		else
			direction *= 0.8f;
		angle = origin.getAngle(TwoFloats(-direction.x, direction.y));

		if(direction.getDistance(desiredDirection) < 0.01f)
			direction = desiredDirection;
	}

	if(direction != 0.0f)
	{
		position += direction * speed;
		collidePosition = position;

		armPos += 0.033f;
		if(armPos > 1.0f)
			armPos -= 1.0f;
	}
	else if(armPos > 0.0f)
	{
		lastArm = armPos;

		armPos += 0.033f;

		if(((lastArm < 0.5f) && (armPos >= 0.5f)) || (armPos > 1.0f))
			armPos = 0.0f;
	}

	if(tail.size() > 0)
	{
		tail.erase(tail.begin());
		tail.push_back(position);
	}

	if(tailFlash)
	{
		--tailFlashCount;
		if(!tailFlashCount)
		{
			ThreeFloats col;

			tailFlashCount = 8;

			if(tailCol[0] == colour)
				col = colour * 0.5f;
			else
				col = colour;

			for(i=0;i<tailCol.size();++i)
				tailCol[i] = col;
		}
	}

	if(powerLevel > 0.87f)
	{
		--flameCount;
		if(!flameCount)
		{
			flameCount = 7;
			++flameIndex;
			flameIndex %= 6;
		}

		if(!(flameCount%3))
		{
			particles.setColour(oldCol + ((newCol-oldCol) * flameInterp));
			particles.trigger(position);
		}

		flameInterp += 0.1f;
		if(flameInterp >= 1.0f)
		{
			ThreeFloats tempCol(newCol);

			newCol = oldCol;
			oldCol = tempCol;

			flameInterp = 0.0f;
		}
	}

	particles.update();
}

//------------------------------------------------------------------------------
void Player::drawTail(shared_ptr<Drawer>& d, float timestepInterp)
{
	if(tail.size() > 4)
	{
		d->drawConnectedLine(L"PlayerTrail",
							 tail,
							 tailSize,
							 tailCol,
							 tailAlpha);
	}
}

//------------------------------------------------------------------------------
void Player::draw(shared_ptr<Drawer>& d, float timestepInterp)
{
	TwoFloats pivot;
	TwoFloats tempPos;
	TwoFloats tempDir(sinf(angle), -cosf(angle));
	TwoFloats perpVector(tempDir.perpendicular());
	float sinArm = sinf(armPos * 2.0f * NIALL_PI);
	const ThreeFloats flameCol(oldCol + ((newCol-oldCol) * flameInterp));

	if(powerLevel > 0.87f)
	{
		d->translate(position);
		d->rotate(angle, 0.0f);

		//Draw flame.
		tempPos.x = -31.0f;
		tempPos.y = -73.0f + flameOffsets[flameIndex];
		d->drawImage(flameImages[flameIndex],
					 tempPos,
					 1.0f,
					 flameCol);

		d->rotate(-angle, 0.0f);
		d->translate(TwoFloats(-position.x, -position.y));
	}

	//Draw particles.
	particles.draw(d, timestepInterp);

	d->translate(position);
	d->rotate(angle, 0.0f);

	//Draw right arm.
	pivot.x = 12.5f;
	pivot.y = 9.5f;
	tempPos.x = -14.0f-11.0f;
	tempPos.y = -2.0f-pivot.y + (sinArm * 2.0f);
	d->drawImage(L"PlayerArmBottom", tempPos);
	d->drawImage(L"PlayerRightArmTop", tempPos, 1.0f, colour);

	//Draw left arm.
	pivot.x = 12.5f;
	pivot.y = 9.5f;
	tempPos.x = -0.0f;
	tempPos.y = -2.0f-pivot.y - (sinArm * 2.0f);
	d->drawImage(L"PlayerArmBottom", tempPos);
	d->drawImage(L"PlayerLeftArmTop", tempPos, 1.0f, colour);

	//Draw body.
	pivot.x = 14.0f;
	pivot.y = 15.0f;
	tempPos.x = -14.0f;
	tempPos.y = -15.0f;
	d->drawImage(L"PlayerBodyBottom", tempPos);
	d->drawImage(L"PlayerBodyTop", tempPos, 1.0f, colour);

	d->rotate(-angle, 0.0f);
	d->translate(TwoFloats(-position.x, -position.y));
}

//------------------------------------------------------------------------------
void Player::reset(const TwoFloats& pos, float ang)
{
	unsigned int i;

	position = pos;
	collidePosition = position;
	angle = ang;
	direction = 0.0f;
	for(i=0;i<tail.size();++i)
		tail[i] = position;
}

//------------------------------------------------------------------------------
void Player::incTail(int numNewSegments)
{
	int i;
	int numToRemove;

	//Remove segments if numNewSegments is negative.
	if(numNewSegments < 0)
	{
		if(tail.size() > 12)
		{
			numToRemove = abs(numNewSegments);

			if((tail.size()-numToRemove) < 12)
				numToRemove = (tail.size()-12);
			for(i=0;i<numToRemove;++i)
			{
				tail.erase(tail.begin());
				tailSize.erase(tailSize.begin());
				tailCol.erase(tailCol.begin());
				tailAlpha.erase(tailAlpha.begin());
			}
		}

		if(tail.size() >= 12)
		{
			for(i=0;i<4;++i)
			{
				tailSize[i] = 0.0f;
				tailAlpha[i] = 0.0f;
			}
			for(i=4;i<12;++i)
				tailSize[i] = fakeLog((float)(i-4)*0.125f)*32.0f;
		}
	}
	//Add segments if numNewSegments is positive.
	else if(numNewSegments > 0)
	{
		const TwoFloats pos(tail[tail.size()-1]);

		//Copy tail's contents to tempTail.
		for(i=0;i<(int)tail.size();++i)
			tempTail[i] = tail[i];

		//Update the drawing vectors.
		for(i=0;i<numNewSegments;++i)
		{
			tail.push_back(pos);
			tempTail.push_front(tail[0]);
			tailSize.push_back(32.0f);
			tailCol.push_back(colour);
			tailAlpha.push_back(1.0f);
		}

		//Copy tempTail's contents (which will be in the correct order) to tail.
		for(i=0;i<(int)tail.size();++i)
			tail[i] = tempTail[i];
	}
}

//------------------------------------------------------------------------------
void Player::setPowerLevel(float val)
{
	int tailLength = StartingTailSize;

	if(val < 0.87f)
		speed = 4.0f + (val * 2.0f);
	else
		speed = 8.0f;
	tailLength += (int)(val * 256.0f);

	if(tail.size() != tailLength)
		incTail(tailLength-tail.size());

	powerLevel = val;
}

//------------------------------------------------------------------------------
void Player::setTailFlash(bool val)
{
	int i;

	if(tailFlash != val)
	{
		tailFlash = val;

		if(!tailFlash)
		{
			for(i=0;i<(int)tailCol.size();++i)
				tailCol[i] = colour;

			tailFlashCount = 1;
		}
	}
}

//------------------------------------------------------------------------------
void Player::keyDown(const std::wstring& action)
{
	if(action == L"Up")
		upIsDown = true;
	else if(action == L"Down")
		downIsDown = true;
	else if(action == L"Left")
		leftIsDown = true;
	else if(action == L"Right")
		rightIsDown = true;

	if(leftIsDown && !rightIsDown)
		desiredDirection.x = -1.0f;
	else if(rightIsDown && ! leftIsDown)
		desiredDirection.x = 1.0f;
	else
		desiredDirection.x = 0.0f;
	if(upIsDown && !downIsDown)
		desiredDirection.y = -1.0f;
	else if(downIsDown && ! upIsDown)
		desiredDirection.y = 1.0f;
	else
		desiredDirection.y = 0.0f;
	if(desiredDirection != 0.0f)
		desiredDirection = desiredDirection.normal();
}

//------------------------------------------------------------------------------
void Player::keyUp(const std::wstring& action)
{
	if(action == L"Up")
		upIsDown = false;
	else if(action == L"Down")
		downIsDown = false;
	else if(action == L"Left")
		leftIsDown = false;
	else if(action == L"Right")
		rightIsDown = false;

	if(leftIsDown && !rightIsDown)
		desiredDirection.x = -1.0f;
	else if(rightIsDown && ! leftIsDown)
		desiredDirection.x = 1.0f;
	else
		desiredDirection.x = 0.0f;
	if(upIsDown && !downIsDown)
		desiredDirection.y = -1.0f;
	else if(downIsDown && ! upIsDown)
		desiredDirection.y = 1.0f;
	else
		desiredDirection.y = 0.0f;
	if(desiredDirection != 0.0f)
		desiredDirection = desiredDirection.normal();
}

//------------------------------------------------------------------------------
void Player::move(const std::wstring& action, float value)
{
	const TwoFloats origin;

	if((action == L"Left") || (action == L"Right"))
	{
		desiredDirection.x = value;
		if(fabsf(desiredDirection.x) < 0.1f)
			desiredDirection.x = 0.0f;
	}
	else if((action == L"Up") || (action == L"Down"))
	{
		desiredDirection.y = value;
		if(fabsf(desiredDirection.y) < 0.1f)
			desiredDirection.y = 0.0f;
	}
}

//------------------------------------------------------------------------------
bool Player::interestedInCollisionsWith(const std::wstring& entity) const
{
	return true;
}

//------------------------------------------------------------------------------
bool Player::collide(const shared_ptr<Collidable>& other, GamePlayState& state)
{
	bool retval = false;

	//if(other->getName() != L"Player")
	{
		float offset;
		const TwoFloats origin;
		TwoFloats diff(other->getPosition() - position);
		float dist = origin.getDistance(diff);
		float rad = getRadius() + other->getRadius();

		diff = diff.normal();
		offset = (rad - dist);
		collidePosition -= diff * offset;
	}

	return retval;
}

//------------------------------------------------------------------------------
float Player::getRadius() const
{
	float retval = 20.5f;

	if(powerLevel >= 0.87f)
		retval = 28.0f;

	return retval;
}
