//	SnakeballSounder.h - The game's sounder.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef SNAKEBALLSOUNDER_H_
#define SNAKEBALLSOUNDER_H_

#include "sound/StreamingSound.h"
#include "sound/Sounder.h"
#include "sound/Delay.h"
#include "Thread.h"

#include <string>
#include <map>

///	The game's sounder.
class SnakeballSounder : public Sounder,
						 public Thread
{
  public:
	///	Constructor.
	SnakeballSounder();
	///	Destructor.
	~SnakeballSounder();
	
	///	Called for every audio block.
	void getAudio(float *input, float *output, int numSamples);

	///	Called for each sound file resource by LoadingScreen.
	void loadSoundFileResource(SoundFileLoaderPtr soundFileLoader);

	///	Used to stream audio off the disk.
	bool run();

	///	Plays a StreamingSound in a loop.
	/*!
		\param sound The name of the sound to play.

		This will tell a StreamingSound to play in a loop until it is replaced
		by another playMusic sound.
	 */
	void playMusic(const std::wstring& sound);
	///	Stops the music playing.
	void stopMusic();
	///	Triggers a StreamingSound to play as a one-shot.
	/*!
		\param The name of the sound to play.
		\param pan How the sound is panned (0->1; left->right).

		This will tell a StreamingSound to play, and stop when it reaches the
		end of its file.
	 */
	void triggerSound(const std::wstring& sound, float level = 1.0f, float pan = 0.5f);
	///	Triggers an input envelope to sound.
	void triggerEnvelope(float level = 1.0f, float pan = 0.5f);
	///	Triggers a sound to loop.
	/*!
		\return The index of the sound. This can be used to stop the sound
		looping with stopLoop().
	 */
	int triggerLoop(const std::wstring& sound, float pan = 0.5f);
	///	Stops the indexed loop playing.
	void stopLoop(int index);

	///	Sets the number of available input channels.
	void setNumChannels(int val);
	///	Sets the red input channel.
	void setRedInputChannel(int val);
	///	Sets the blue input channel.
	void setBlueInputChannel(int val);

	///	If we're playing Shouting Snakeball, no delay on enveloped input audio.
	void setShouting(bool val);

	///	Returns the red input level.
	float getRedInput() const {return redInput;};
	///	Returns the blue input level.
	float getBlueInput() const {return blueInput;};

	///	Returns true if the music track's already playing.
	bool getMusicIsPlaying() const {return (music > -1);};
  private:
	///	Struct representing a pre-loaded buffer.
	struct PreLoadedBuffer
	{
		///	Pointer to the buffer data.
		float *buffer;
		///	The size of the buffer data.
		int size;
	};
	///	Struct representing a single enveloped section of the input audio.
	struct InputEnvelope
	{
		///	The index of the envelope (0->1).
		float index;
		///	Used to ramp the envelope up & down (-1->1).
		float inc;
		///	The maximum level value.
		float level;
		///	The left channel amount (envelope value gets multiplied by this).
		float left;
		///	The right channel amount (envelope value gets multiplied by this).
		float right;
	};

	///	The number of available StreamingSounds.
	enum
	{
		NumSounds = 128,
		InputBufferSize = 44100
	};

	///	Map of sound file names and their associated SoundFileLoaders.
	/*!
		Used for longer files.
	 */
	std::map<std::wstring, SoundFileLoaderPtr> soundFiles;
	///	Map of sound file names and their pre-loaded buffers.
	/*!
		Used for short files, which can reasonably be read entirely into
		memory.
	 */
	std::map<std::wstring, PreLoadedBuffer> preLoadedBuffers;

	///	The available StreamingSounds.
	StreamingSound sounds[NumSounds];
	///	The available InputEnvelopes.
	InputEnvelope envelopes[NumSounds];

	///	The input audio buffer.
	float inputBuffer[2][InputBufferSize];
	///	Our current write position within the input buffer.
	/*!
		(our read position is this + 1 i.e. the length of the buffer away)
	 */
	int inputPos;

	///	Index of the current music StreamingSound.
	int music;

	///	Used to kill the thread.
	bool timeToStop;

	///	The delay applied to the sound files.
	Delay delay;

	///	The number of available input channels.
	int numChannels;
	///	The red input channel.
	int redInputChannel;
	///	The blue input channel.
	int blueInputChannel;
	///	The red input level.
	float redInput;
	///	The blue input level.
	float blueInput;

	///	True if we're playing Shouting Snakeball.
	bool shouting;
};

#endif
